import re
import socket
from typing import List, Optional

import pytest

from .env import H1Conf


class TestRequestStrict:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        conf = H1Conf(env)
        conf.add([
            "HttpProtocolOptions Strict",
        ])
        conf.install()
        assert env.apache_restart() == 0

    # strict tests from t/apache/http_strict.t
    # possible expected results:
    #   0:       any HTTP error
    #   1:       any HTTP success
    #   200-500: specific HTTP status code
    #   undef:   HTTPD should drop connection without error message
    @pytest.mark.parametrize(["intext", "status"], [
        ["GET / HTTP/1.0\n\n", 400],
        ["G/T / HTTP/1.0\r\n\r\n", 400],
        ["GET / HTTP/1.0  \r\nHost: localhost\r\n\r\n", 400],
        ["GET / HTTP/1.0\r\nFoo: b\x01ar\r\n\r\n", 400],
        ["GET / HTTP/1.0\r\nF\x01o: bar\r\n\r\n", 400],
        ["GET / HTTP/1.0\r\r", None],
        ["GET / HTTP/1.0\r\nHost: localhost\r\nHost: localhost\r\n\r\n", 400],
        ["GET http://017700000001/ HTTP/1.0\r\n\r\n", 400],
        ["GET http://0x7f.1/ HTTP/1.0\r\n\r\n", 400],
        ["GET http://127.01.0.1/ HTTP/1.0\r\n\r\n", 400],
        ["GET http://%3127.0.0.1/ HTTP/1.0\r\n\r\n", 400],
        ["GET / HTTP/1.0\r\nHost: localhost:80\r\nHost: localhost:80\r\n\r\n", 400],
        ["GET http://foo@localhost:80/ HTTP/1.0\r\n\r\n", 400],
        ["GET / HTTP/1.0\r\nHost: 4714::abcd:8001\r\n\r\n", 400],
        ["GET / HTTP/1.0\r\nHost: abc\xa0\r\n\r\n", 400],
        ["GET / HTTP/1.0\r\nHost: foo_bar.example.com\r\n\r\n", 200],
        ["GET http://foo_bar/ HTTP/1.0\r\n\r\n", 200],
    ])
    def test_h1_007_01(self, env, intext, status: Optional[int]):
        with socket.create_connection(('localhost', int(env.http_port))) as sock:
            # on some OS, the server does not see our connection until there is
            # something incoming
            sock.sendall(intext.encode())
            sock.shutdown(socket.SHUT_WR)
            buff = sock.recv(1024)
            msg = buff.decode()
            if status is None:
                assert len(msg) == 0, f"unexpected answer: {msg}"
            else:
                assert len(msg) > 0, "no answer from server"
                rlines = msg.splitlines()
                response = rlines[0]
                m = re.match(r'^HTTP/1.1 (\d+)\s+(\S+)', response)
                assert m, f"unrecognized response: {rlines}"
                if status == 1:
                    assert int(m.group(1)) >= 200
                elif status == 90:
                    assert len(rlines) >= 1, f"{rlines}"
                elif status > 0:
                    assert int(m.group(1)) == status, f"{rlines}"
                else:
                    assert int(m.group(1)) >= 400, f"{rlines}"

    @pytest.mark.parametrize(["hvalue", "expvalue", "status", "lognos"], [
        ['"123"', '123', 200, None],
        ['"123 "', '123 ', 200, None],       # trailing space stays
        ['"123\t"', '123\t', 200, None],     # trailing tab stays
        ['" 123"', '123', 200, None],        # leading space is stripped
        ['"          123"', '123', 200, None],  # leading spaces are stripped
        ['"\t123"', '123', 200, None],       # leading tab is stripped
        ['"expr=%{unescape:123%0A 123}"', '', 500, ["AH02430"]],  # illegal char
        ['" \t "', '', 200, None],           # just ws
    ])
    def test_h1_007_02(self, env, hvalue, expvalue, status, lognos: Optional[List[str]]):
        hname = 'ap-test-007'
        conf = H1Conf(env, extras={
            f'test1.{env.http_tld}': [
                '<Location /index.html>',
                f'Header add {hname} {hvalue}',
                '</Location>',
            ]
        })
        conf.add_vhost_test1(proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "test1", "/index.html")
        r = env.curl_get(url, options=['--http1.1'])
        assert r.response["status"] == status
        if int(status) < 400:
            assert r.response["header"][hname] == expvalue
        #
        if lognos is not None:
            env.httpd_error_log.ignore_recent(lognos = lognos)

    @pytest.mark.parametrize(["hvalue", "expvalue"], [
        ['123', '123'],
        ['123 ', '123'],    # trailing space is stripped
        ['123\t', '123'],    # trailing tab is stripped
        [' 123', '123'],    # leading space is stripped
        ['          123', '123'],  # leading spaces are stripped
        ['\t123', '123'],  # leading tab is stripped
    ])
    def test_h1_007_03(self, env, hvalue, expvalue):
        # same as 007_02, but http1 proxied
        hname = 'ap-test-007'
        conf = H1Conf(env, extras={
            f'test1.{env.http_tld}': [
                '<Location /index.html>',
                f'Header add {hname} "{hvalue}"',
                '</Location>',
            ]
        })
        conf.add_vhost_test1(proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "test1", "/proxy/index.html")
        r = env.curl_get(url, options=['--http1.1'])
        assert r.response["status"] == 200
        assert r.response["header"][hname] == expvalue



import pytest
from typing import List, Optional

from pyhttpd.env import HttpdTestEnv
from .env import H2Conf, H2TestEnv


@pytest.mark.skipif(condition=H2TestEnv.is_unsupported, reason="mod_http2 not supported here")
class TestRfc9113:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        H2Conf(env).add_vhost_test1().install()
        assert env.apache_restart() == 0

    # by default, we accept leading/trailing ws in request fields
    def test_h2_203_01_ws_ignore(self, env):
        url = env.mkurl("https", "test1", "/")
        r = env.curl_get(url, options=['-H', 'trailing-space: must not  '])
        assert r.exit_code == 0, f'curl output: {r.stderr}'
        assert r.response["status"] == 200, f'curl output: {r.stdout}'
        r = env.curl_get(url, options=['-H', 'trailing-space: must not\t'])
        assert r.exit_code == 0, f'curl output: {r.stderr}'
        assert r.response["status"] == 200, f'curl output: {r.stdout}'

    # response header are also handled, but we strip ws before sending
    @pytest.mark.parametrize(["hvalue", "expvalue", "status", "lognos"], [
        ['"123"', '123', 200, None],
        ['"123 "', '123', 200, None],       # trailing space stripped
        ['"123\t"', '123', 200, None],     # trailing tab stripped
        ['" 123"', '123', 200, None],        # leading space is stripped
        ['"          123"', '123', 200, None],  # leading spaces are stripped
        ['"\t123"', '123', 200, None],       # leading tab is stripped
        ['"expr=%{unescape:123%0A 123}"', '', 500, ["AH02430"]],  # illegal char
        ['" \t "', '', 200, None],          # just ws
    ])
    def test_h2_203_02(self, env, hvalue, expvalue, status, lognos: Optional[List[str]]):
        hname = 'ap-test-007'
        conf = H2Conf(env, extras={
            f'test1.{env.http_tld}': [
                '<Location /index.html>',
                f'Header add {hname} {hvalue}',
                '</Location>',
            ]
        })
        conf.add_vhost_test1(proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "test1", "/index.html")
        r = env.curl_get(url, options=['--http2'])
        if status == 500 and r.exit_code != 0:
            # in 2.4.x we fail late on control chars in a response
            # and RST_STREAM. That's also ok
            return
        assert r.response["status"] == status
        if int(status) < 400:
            assert r.response["header"][hname] == expvalue
        #
        if lognos is not None:
            env.httpd_error_log.ignore_recent(lognos = lognos)




import socket

import pytest

from .env import H1Conf


class TestGet:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        H1Conf(env).add_vhost_cgi(
            proxy_self=True
        ).add_vhost_test1(
            proxy_self=True
        ).install()
        assert env.apache_restart() == 0

    # check SSL environment variables from CGI script
    def test_h1_003_01(self, env):
        url = env.mkurl("https", "cgi", "/hello.py")
        r = env.curl_get(url)
        assert r.response["status"] == 200
        assert r.response["json"]["protocol"] == "HTTP/1.1"
        assert r.response["json"]["https"] == "on"
        tls_version = r.response["json"]["ssl_protocol"]
        assert tls_version in ["TLSv1.2", "TLSv1.3"]



import datetime
import email.utils
import os
from datetime import timedelta

import pytest
from pyhttpd.certs import CertificateSpec

from pyhttpd.env import HttpdTestEnv
from .md_cert_util import MDCertUtil
from .md_env import MDTestEnv
from .md_conf import MDConf


@pytest.mark.skipif(condition=not MDTestEnv.has_acme_server(),
                    reason="no ACME test server configured")
class TestProfiles:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env, acme):
        env.APACHE_CONF_SRC = "data/test_auto"
        acme.start(config='default')
        env.check_acme()
        env.clear_store()
        MDConf(env).install()
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'

    @pytest.fixture(autouse=True, scope='function')
    def _method_scope(self, env, request):
        env.clear_store()
        self.test_domain = env.get_request_domain(request)

    def _write_res_file(self, doc_root, name, content):
        if not os.path.exists(doc_root):
            os.makedirs(doc_root)
        open(os.path.join(doc_root, name), "w").write(content)

    # create a MD with 'default' profile, get cert
    def test_md_710_001(self, env):
        domain = self.test_domain
        # generate config with one MD
        domains = [domain, "www." + domain]
        conf = MDConf(env, admin="admin@" + domain)
        conf.add_drive_mode("auto")
        conf.start_md(domains)
        conf.add(f'  MDProfile default')
        conf.end_md()
        conf.add_vhost(domains)
        conf.install()
        #
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        assert env.await_completion(domains)
        stat = env.get_md_status(domain)
        assert stat["watched"] == 1
        assert stat["profile"] == "default", f'{stat}'
        assert stat['cert']['rsa']['valid']['until'], f'{stat}'
        ts = email.utils.parsedate_to_datetime(stat['cert']['rsa']['valid']['until'])
        valid = ts - datetime.datetime.now(datetime.timezone.utc)
        assert valid.days in [89, 90]

    # create a MD with 'shortlived' profile, get cert
    def test_md_710_002(self, env):
        domain = self.test_domain
        # generate config with one MD
        domains = [domain, "www." + domain]
        conf = MDConf(env, admin="admin@" + domain)
        conf.add_drive_mode("auto")
        conf.start_md(domains)
        conf.add(f'  MDProfile shortlived')
        conf.add(f'  MDProfileMandatory on')
        conf.end_md()
        conf.add_vhost(domains)
        conf.install()
        #
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        assert env.await_completion(domains)
        stat = env.get_md_status(domain)
        assert stat["watched"] == 1
        assert stat["profile"] == "shortlived", f'{stat}'
        assert stat['cert']['rsa']['valid']['until'], f'{stat}'
        ts = email.utils.parsedate_to_datetime(stat['cert']['rsa']['valid']['until'])
        valid = ts - datetime.datetime.now(datetime.timezone.utc)
        assert valid.days in [5, 6]

    # create a MD with unknown 'XXX' profile, get cert
    def test_md_710_003(self, env):
        domain = self.test_domain
        # generate config with one MD
        domains = [domain, "www." + domain]
        conf = MDConf(env, admin="admin@" + domain)
        conf.add_drive_mode("auto")
        conf.start_md(domains)
        conf.add(f'  MDProfile XXX')
        conf.end_md()
        conf.add_vhost(domains)
        conf.install()
        #
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        assert env.await_completion(domains)
        stat = env.get_md_status(domain)
        assert stat["watched"] == 1
        assert stat["profile"] == "XXX", f'{stat}'

    # create a MD with unknown 'XXX' profile, mandatory, fail
    def test_md_710_004(self, env):
        domain = self.test_domain
        # generate config with one MD
        domains = [domain, "www." + domain]
        conf = MDConf(env, admin="admin@" + domain)
        conf.add_drive_mode("auto")
        conf.start_md(domains)
        conf.add(f'  MDProfile XXX')
        conf.add(f'  MDProfileMandatory on')
        conf.end_md()
        conf.add_vhost(domains)
        conf.install()
        #
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        assert env.await_error(domain)
        stat = env.get_md_status(domain)
        assert stat["watched"] == 1
        assert stat["profile"] == "XXX", f'{stat}'
        assert len(stat['cert']) == 0, f'{stat}'
        assert stat['renewal']['errors'] > 0, f'{stat}'
        assert stat['renewal']['last']['activity'] == 'Creating new order, key-spec=default, profile=XXX, replacing-cert=none', f'{stat}'
        MDConf(env).install()
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        env.httpd_error_log.ignore_recent(matches=[
            r'.*mandatory ACME profile \'XXX\' is not offered by CA.*',
        ], lognos=[
            "AH10056"  # processing failed
        ])



#!/usr/bin/env python3
import os, sys
from requestparser import get_request_params


forms, files = get_request_params()
text = forms['text'] if 'text' in forms else "a"
count = int(forms['count']) if 'count' in forms else 77784

print("Status: 200 OK")
print("Content-Type: text/html")
print()
sys.stdout.flush()
for _ in range(count):
    sys.stdout.write(text)




import inspect
import logging
import os

from pyhttpd.env import HttpdTestEnv, HttpdTestSetup

log = logging.getLogger(__name__)


class CoreTestSetup(HttpdTestSetup):

    def __init__(self, env: 'HttpdTestEnv'):
        super().__init__(env=env)
        self.add_source_dir(os.path.dirname(inspect.getfile(CoreTestSetup)))
        self.add_modules(["cgid"])


class CoreTestEnv(HttpdTestEnv):

    def __init__(self, pytestconfig=None):
        super().__init__(pytestconfig=pytestconfig)
        self.add_httpd_log_modules(["http", "core"])

    def setup_httpd(self, setup: HttpdTestSetup = None):
        super().setup_httpd(setup=CoreTestSetup(env=self))



import os

import pytest
from filelock import Timeout, FileLock

from .md_cert_util import MDCertUtil
from .md_conf import MDConf
from .md_env import MDTestEnv


@pytest.mark.skipif(condition=not MDTestEnv.has_acme_server(),
                    reason="no ACME test server configured")
class TestLocks:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env, acme):
        env.APACHE_CONF_SRC = "data/test_auto"
        acme.start(config='default')
        env.check_acme()
        env.clear_store()

    @pytest.fixture(autouse=True, scope='function')
    def _method_scope(self, env, request):
        env.clear_store()
        self.test_domain = env.get_request_domain(request)

    def configure_httpd(self, env, domains, add_lines=""):
        conf = MDConf(env)
        conf.add(add_lines)
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()

    # normal renewal with store locks activated
    def test_md_820_001(self, env):
        domain = self.test_domain
        self.configure_httpd(env, [domain], add_lines=[
            "MDStoreLocks 1s"
        ])
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        assert env.await_completion([domain])

    # renewal, with global lock held during restert
    @pytest.mark.skip("does not work in our CI")
    def test_md_820_002(self, env):
        domain = self.test_domain
        self.configure_httpd(env, [domain], add_lines=[
            "MDStoreLocks 1s"
        ])
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        assert env.await_completion([domain])
        # we have a cert now, add a dns name to force renewal
        certa = MDCertUtil(env.store_domain_file(domain, 'pubcert.pem'))
        self.configure_httpd(env, [domain, f"x.{domain}"], add_lines=[
            "MDStoreLocks 1s"
        ])
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        # await new cert, but do not restart, keeps the cert in staging
        assert env.await_completion([domain], restart=False)
        # obtain global lock and restart
        lockfile = os.path.join(env.store_dir, "store.lock")
        with FileLock(lockfile):
            assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        # lock should have prevented staging from being activated,
        # meaning we will have the same cert
        certb = MDCertUtil(env.store_domain_file(domain, 'pubcert.pem'))
        assert certa.same_serial_as(certb)
        # now restart without lock
        assert env.apache_restart() == 0, f'{env.apachectl_stderr}'
        certc = MDCertUtil(env.store_domain_file(domain, 'pubcert.pem'))
        assert not certa.same_serial_as(certc)





#!/usr/bin/env python3
import os, sys
from requestparser import get_request_params


forms, files = get_request_params()
name = forms['name'] if 'name' in forms else None

if name:
    print("Status: 200")
    print("""\
Content-Type: text/plain\n""")
    print("""%s: %s""" % (name, os.environ['HTTP_'+name]))
else:
    print("Status: 400 Parameter Missing")
    print("""\
Content-Type: text/html\n
<html><body>
<p>No name was specified</p>
</body></html>""")





import os
import re
import time
from datetime import datetime, timedelta
from threading import Thread

import pytest

from .env import CoreTestEnv
from pyhttpd.conf import HttpdConf


class Loader:

    def __init__(self, env, url: str, clients: int, req_per_client: int = 10):
        self.env = env
        self.url = url
        self.clients = clients
        self.req_per_client = req_per_client
        self.result = None
        self.total_request = 0
        self._thread = None

    def run(self):
        self.total_requests = self.clients * self.req_per_client
        conn_per_client = 5
        args = [self.env.h2load, f"--connect-to=localhost:{self.env.https_port}",
                "--h1",                                 # use only http/1.1
                "-n", str(self.total_requests),              # total # of requests to make
                "-c", str(conn_per_client * self.clients),   # total # of connections to make
                "-r", str(self.clients),                     # connections at a time
                "--rate-period", "2",                   # create conns every 2 sec
                self.url,
                ]
        self.result = self.env.run(args)

    def start(self):
        self._thread = Thread(target=self.run)
        self._thread.start()

    def join(self):
        self._thread.join()


class ChildDynamics:

    RE_DATE_TIME = re.compile(r'\[(?P<date_time>[^\]]+)\] .*')
    RE_TIME_FRAC = re.compile(r'(?P<dt>.* \d\d:\d\d:\d\d)(?P<frac>.(?P<micros>.\d+)) (?P<year>\d+)')
    RE_CHILD_CHANGE = re.compile(r'\[(?P<date_time>[^\]]+)\] '
                                 r'\[mpm_event:\w+\]'
                                 r' \[pid (?P<main_pid>\d+):tid \w+\] '
                                 r'.* Child (?P<child_no>\d+) (?P<action>\w+): '
                                 r'pid (?P<pid>\d+), gen (?P<generation>\d+), .*')

    def __init__(self, env: CoreTestEnv):
        self.env = env
        self.changes = list()
        self._start = None
        for l in open(env.httpd_error_log.path):
            m = self.RE_CHILD_CHANGE.match(l)
            if m:
                self.changes.append({
                    'pid': int(m.group('pid')),
                    'child_no': int(m.group('child_no')),
                    'gen': int(m.group('generation')),
                    'action': m.group('action'),
                    'rtime' : self._rtime(m.group('date_time'))
                })
                continue
            if self._start is None:
                m = self.RE_DATE_TIME.match(l)
                if m:
                    self._rtime(m.group('date_time'))

    def _rtime(self, s: str) -> timedelta:
        micros = 0
        m = self.RE_TIME_FRAC.match(s)
        if m:
            micros = int(m.group('micros'))
            s = f"{m.group('dt')} {m.group('year')}"
        d = datetime.strptime(s, '%a %b %d %H:%M:%S %Y') + timedelta(microseconds=micros)
        if self._start is None:
            self._start = d
        delta = d - self._start
        return f"{delta.seconds:+02d}.{delta.microseconds:06d}"



@pytest.mark.skipif(condition='STRESS_TEST' not in os.environ,
                    reason="STRESS_TEST not set in env")
@pytest.mark.skipif(condition=not CoreTestEnv().h2load_is_at_least('1.41.0'),
                    reason="h2load unavailable or misses --connect-to option")
class TestRestarts:

    def test_core_002_01(self, env):
        # Lets make a tight config that triggers dynamic child behaviour
        conf = HttpdConf(env, extras={
            'base': f"""
        StartServers            1
        ServerLimit             3
        ThreadLimit             4
        ThreadsPerChild         4
        MinSpareThreads         4
        MaxSpareThreads         6
        MaxRequestWorkers       12
        MaxConnectionsPerChild  0

        LogLevel mpm_event:trace6
                """,
        })
        conf.add_vhost_cgi()
        conf.install()

        # clear logs and start server, start load
        env.httpd_error_log.clear_log()
        assert env.apache_restart() == 0
        # we should see a single child started
        cd = ChildDynamics(env)
        assert len(cd.changes) == 1, f"{cd.changes}"
        assert cd.changes[0]['action'] == 'started'
        # This loader simulates 6 clients, each making 10 requests.
        # delay.py sleeps for 1sec, so this should run for about 10 seconds
        loader = Loader(env=env, url=env.mkurl("https", "cgi", "/delay.py"),
                        clients=6, req_per_client=10)
        loader.start()
        # Expect 2 more children to have been started after half time
        time.sleep(5)
        cd = ChildDynamics(env)
        assert len(cd.changes) == 3, f"{cd.changes}"
        assert len([x for x in cd.changes if x['action'] == 'started']) == 3, f"{cd.changes}"

        # Trigger a server reload
        assert env.apache_reload() == 0
        # a graceful reload lets ongoing requests continue, but
        # after a while all gen 0 children should have stopped
        time.sleep(3)  # FIXME: this pbly depends on the runtime a lot, do we have expectations?
        cd = ChildDynamics(env)
        gen0 = [x for x in cd.changes if x['gen'] == 0]
        assert len([x for x in gen0 if x['action'] == 'stopped']) == 3

        # wait for the loader to finish and stop the server
        loader.join()
        env.apache_stop()

        # Similar to before the reload, we expect 3 children to have
        # been started and stopped again on server stop
        cd = ChildDynamics(env)
        gen1 = [x for x in cd.changes if x['gen'] == 1]
        assert len([x for x in gen1 if x['action'] == 'started']) == 3
        assert len([x for x in gen1 if x['action'] == 'stopped']) == 3



#!/usr/bin/env python3

import os
import json

resp = {
    'https': os.getenv('HTTPS', ''),
    'host': os.getenv('X_HOST', '') if 'X_HOST' in os.environ else os.getenv('SERVER_NAME', ''),
    'server': os.getenv('SERVER_NAME', ''),
    'h2_original_host': os.getenv('H2_ORIGINAL_HOST', ''),
    'port': os.getenv('SERVER_PORT', ''),
    'protocol': os.getenv('SERVER_PROTOCOL', ''),
    'ssl_protocol': os.getenv('SSL_PROTOCOL', ''),
    'h2': os.getenv('HTTP2', ''),
    'h2push': os.getenv('H2PUSH', ''),
    'h2_stream_id': os.getenv('H2_STREAM_ID', ''),
    'x-forwarded-for': os.getenv('HTTP_X_FORWARDED_FOR', ''),
    'x-forwarded-host': os.getenv('HTTP_X_FORWARDED_HOST', ''),
    'x-forwarded-server': os.getenv('HTTP_X_FORWARDED_SERVER', ''),
}

print("Content-Type: application/json")
print()
print(json.JSONEncoder(indent=2).encode(resp))




import pytest

from .env import H2Conf, H2TestEnv


@pytest.mark.skipif(condition=H2TestEnv.is_unsupported, reason="mod_http2 not supported here")
class TestH2Proxy:

    def test_h2_600_01(self, env):
        conf = H2Conf(env, extras={
            f'cgi.{env.http_tld}': [
                "SetEnvIf Host (.+) X_HOST=$1",
            ]
        })
        conf.add_vhost_cgi(h2proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", "/h2proxy/hello.py")
        r = env.curl_get(url, 5)
        assert r.response["status"] == 200
        assert r.response["json"]["protocol"] == "HTTP/2.0"
        assert r.response["json"]["https"] == "on"
        assert r.response["json"]["ssl_protocol"] != ""
        assert r.response["json"]["h2"] == "on"
        assert r.response["json"]["h2push"] == "off"
        assert r.response["json"]["host"] == f"cgi.{env.http_tld}:{env.https_port}"

    def test_h2_600_02(self, env):
        conf = H2Conf(env, extras={
            f'cgi.{env.http_tld}': [
                "SetEnvIf Host (.+) X_HOST=$1",
                f"ProxyPreserveHost on",
                f"ProxyPass /h2c/ h2c://127.0.0.1:{env.http_port}/",
            ]
        })
        conf.add_vhost_cgi()
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", "/h2c/hello.py")
        r = env.curl_get(url, 5)
        assert r.response["status"] == 200
        assert r.response["json"]["protocol"] == "HTTP/2.0"
        assert r.response["json"]["https"] == ""
        # the proxied backend sees Host header as passed on front
        assert r.response["json"]["host"] == f"cgi.{env.http_tld}:{env.https_port}"
        assert r.response["json"]["h2_original_host"] == ""

    def test_h2_600_03(self, env):
        conf = H2Conf(env, extras={
            f'cgi.{env.http_tld}': [
                "SetEnvIf Host (.+) X_HOST=$1",
                f"ProxyPreserveHost off",
                f"ProxyPass /h2c/ h2c://127.0.0.1:{env.http_port}/",
            ]
        })
        conf.add_vhost_cgi()
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", "/h2c/hello.py")
        r = env.curl_get(url, 5)
        assert r.response["status"] == 200
        assert r.response["json"]["protocol"] == "HTTP/2.0"
        assert r.response["json"]["https"] == ""
        # the proxied backend sees Host as using in connecting to it
        assert r.response["json"]["host"] == f"127.0.0.1:{env.http_port}"
        assert r.response["json"]["h2_original_host"] == ""

    # check that connection reuse actually happens as configured
    @pytest.mark.parametrize("enable_reuse", [ "on", "off" ])
    def test_h2_600_04(self, env, enable_reuse):
        conf = H2Conf(env, extras={
            f'cgi.{env.http_tld}': [
                f"ProxyPassMatch ^/h2proxy/([0-9]+)/(.*)$ "
                f"  h2c://127.0.0.1:$1/$2 enablereuse={enable_reuse} keepalive=on",
            ]
        })
        conf.add_vhost_cgi()
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", f"/h2proxy/{env.http_port}/hello.py")
        # httpd 2.5.0 disables reuse, not matter the config
        if enable_reuse == "on" and not env.httpd_is_at_least("2.4.60"):
            # reuse is not guaranteed for each request, but we expect some
            # to do it and run on a h2 stream id > 1
            reused = False
            count = 10
            r = env.curl_raw([url] * count, 5)
            response = r.response
            for n in range(count):
                assert response["status"] == 200
                if n == (count - 1):
                    break
                response = response["previous"]
            assert r.json[0]["h2_stream_id"] == "1"
            for n in range(1, count):
                if int(r.json[n]["h2_stream_id"]) > 1:
                    reused = True
                    break
            assert reused
        else:
            r = env.curl_raw([url, url], 5)
            assert r.response["previous"]["status"] == 200
            assert r.response["status"] == 200
            assert r.json[0]["h2_stream_id"] == "1"
            assert r.json[1]["h2_stream_id"] == "1"

    # do some flexible setup from #235 to proper connection selection
    @pytest.mark.parametrize("enable_reuse", [ "on", "off" ])
    def test_h2_600_05(self, env, enable_reuse):
        conf = H2Conf(env, extras={
            f'cgi.{env.http_tld}': [
                f"ProxyPassMatch ^/h2proxy/([0-9]+)/(.*)$ "
                f"  h2c://127.0.0.1:$1/$2 enablereuse={enable_reuse} keepalive=on",
            ]
        })
        conf.add_vhost_cgi()
        conf.add([
            f'Listen {env.http_port2}',
            'UseCanonicalName On',
            'UseCanonicalPhysicalPort On'
        ])
        conf.start_vhost(domains=[f'cgi.{env.http_tld}'],
                         port=5004, doc_root="htdocs/cgi")
        conf.add("AddHandler cgi-script .py")
        conf.end_vhost()
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", f"/h2proxy/{env.http_port}/hello.py")
        url2 = env.mkurl("https", "cgi", f"/h2proxy/{env.http_port2}/hello.py")
        r = env.curl_raw([url, url2], 5)
        assert r.response["previous"]["status"] == 200
        assert int(r.json[0]["port"]) == env.http_port
        assert r.response["status"] == 200
        exp_port = env.http_port if enable_reuse == "on" \
                                    and not env.httpd_is_at_least("2.4.60")\
            else env.http_port2
        assert int(r.json[1]["port"]) == exp_port

    # test X-Forwarded-* headers
    def test_h2_600_06(self, env):
        conf = H2Conf(env, extras={
            f'cgi.{env.http_tld}': [
                "SetEnvIf Host (.+) X_HOST=$1",
                f"ProxyPreserveHost on",
                f"ProxyPass /h2c/ h2c://127.0.0.1:{env.http_port}/",
                f"ProxyPass /h1c/ http://127.0.0.1:{env.http_port}/",
            ]
        })
        conf.add_vhost_cgi(proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", "/h1c/hello.py")
        r1 = env.curl_get(url, 5)
        assert r1.response["status"] == 200
        url = env.mkurl("https", "cgi", "/h2c/hello.py")
        r2 = env.curl_get(url, 5)
        assert r2.response["status"] == 200
        for key in ['x-forwarded-for', 'x-forwarded-host','x-forwarded-server']:
            assert r1.json[key] == r2.json[key], f'{key} differs proxy_http != proxy_http2'

    # lets do some error tests
    def test_h2_600_30(self, env):
        conf = H2Conf(env)
        conf.add_vhost_cgi(h2proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", "/h2proxy/h2test/error?status=500")
        r = env.curl_get(url)
        assert r.exit_code == 0, r
        assert r.response['status'] == 500
        url = env.mkurl("https", "cgi", "/h2proxy/h2test/error?error=timeout")
        r = env.curl_get(url)
        assert r.exit_code == 0, r
        assert r.response['status'] == 408

    # produce an error during response body
    def test_h2_600_31(self, env, repeat):
        conf = H2Conf(env)
        conf.add_vhost_cgi(h2proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", "/h2proxy/h2test/error?body_error=timeout")
        r = env.curl_get(url)
        # depending on when the error is detect in proxying, if may RST the
        # stream (exit_code != 0) or give a 503 response.
        if r.exit_code == 0:
            assert r.response['status'] == 502

    # produce an error, fail to generate an error bucket
    def test_h2_600_32(self, env, repeat):
        conf = H2Conf(env)
        conf.add_vhost_cgi(h2proxy_self=True)
        conf.install()
        assert env.apache_restart() == 0
        url = env.mkurl("https", "cgi", "/h2proxy/h2test/error?body_error=timeout&error_bucket=0")
        r = env.curl_get(url)
        # depending on when the error is detect in proxying, if may RST the
        # stream (exit_code != 0) or give a 503 response.
        if r.exit_code == 0:
            assert r.response['status'] in [502, 503]



import inspect
import logging
import os
import subprocess
from typing import Dict, Any

from pyhttpd.certs import CertificateSpec
from pyhttpd.conf import HttpdConf
from pyhttpd.env import HttpdTestEnv, HttpdTestSetup

log = logging.getLogger(__name__)


class H1TestSetup(HttpdTestSetup):

    def __init__(self, env: 'HttpdTestEnv'):
        super().__init__(env=env)
        self.add_source_dir(os.path.dirname(inspect.getfile(H1TestSetup)))
        self.add_modules(["cgid", "autoindex", "ssl"])

    def make(self):
        super().make()
        self._add_h1test()
        self._setup_data_1k_1m()

    def _add_h1test(self):
        local_dir = os.path.dirname(inspect.getfile(H1TestSetup))
        p = subprocess.run([self.env.apxs, '-c', 'mod_h1test.c'],
                           capture_output=True,
                           cwd=os.path.join(local_dir, 'mod_h1test'))
        rv = p.returncode
        if rv != 0:
            log.error(f"compiling md_h1test failed: {p.stderr}")
            raise Exception(f"compiling md_h1test failed: {p.stderr}")

        modules_conf = os.path.join(self.env.server_dir, 'conf/modules.conf')
        with open(modules_conf, 'a') as fd:
            # load our test module which is not installed
            fd.write(f"LoadModule h1test_module   \"{local_dir}/mod_h1test/.libs/mod_h1test.so\"\n")

    def _setup_data_1k_1m(self):
        s90 = "01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678\n"
        with open(os.path.join(self.env.gen_dir, "data-1k"), 'w') as f:
            for i in range(10):
                f.write(f"{i:09d}-{s90}")
        with open(os.path.join(self.env.gen_dir, "data-10k"), 'w') as f:
            for i in range(100):
                f.write(f"{i:09d}-{s90}")
        with open(os.path.join(self.env.gen_dir, "data-100k"), 'w') as f:
            for i in range(1000):
                f.write(f"{i:09d}-{s90}")
        with open(os.path.join(self.env.gen_dir, "data-1m"), 'w') as f:
            for i in range(10000):
                f.write(f"{i:09d}-{s90}")


class H1TestEnv(HttpdTestEnv):

    def __init__(self, pytestconfig=None):
        super().__init__(pytestconfig=pytestconfig)
        self.add_httpd_log_modules(["http", "core"])

    def setup_httpd(self, setup: HttpdTestSetup = None):
        super().setup_httpd(setup=H1TestSetup(env=self))


class H1Conf(HttpdConf):

    def __init__(self, env: HttpdTestEnv, extras: Dict[str, Any] = None):
        super().__init__(env=env, extras=HttpdConf.merge_extras(extras, {
            "base": [
                "LogLevel http:trace4",
            ],
            f"cgi.{env.http_tld}": [
                "SSLOptions +StdEnvVars",
                "AddHandler cgi-script .py",
                "<Location \"/h1test/echo\">",
                "    SetHandler h1test-echo",
                "</Location>",
            ]
        }))



import logging
import struct


log = logging.getLogger(__name__)


class WsFrame:

    CONT = 0
    TEXT = 1
    BINARY = 2
    RSVD3 = 3
    RSVD4 = 4
    RSVD5 = 5
    RSVD6 = 6
    RSVD7 = 7
    CLOSE = 8
    PING = 9
    PONG = 10
    RSVD11 = 11
    RSVD12 = 12
    RSVD13 = 13
    RSVD14 = 14
    RSVD15 = 15

    OP_NAMES = [
        "CONT",
        "TEXT",
        "BINARY",
        "RSVD3",
        "RSVD4",
        "RSVD5",
        "RSVD6",
        "RSVD7",
        "CLOSE",
        "PING",
        "PONG",
        "RSVD11",
        "RSVD12",
        "RSVD13",
        "RSVD14",
        "RSVD15",
    ]

    def __init__(self, opcode: int, fin: bool, mask: bytes, data: bytes):
        self.opcode = opcode
        self.fin = fin
        self.mask = mask
        self.data = data
        self.length = len(data)

    def __repr__(self):
        return f'WsFrame[{self.OP_NAMES[self.opcode]} fin={self.fin}, mask={self.mask}, len={len(self.data)}]'

    @property
    def data_len(self) -> int:
        return len(self.data) if self.data else 0

    def to_network(self) -> bytes:
        nd = bytearray()
        h1 = self.opcode
        if self.fin:
            h1 |= 0x80
        nd.extend(struct.pack("!B", h1))
        mask_bit = 0x80 if self.mask is not None else 0x0
        h2 = self.data_len
        if h2 > 65535:
            nd.extend(struct.pack("!BQ", 127|mask_bit, h2))
        elif h2 > 126:
            nd.extend(struct.pack("!BH", 126|mask_bit, h2))
        else:
            nd.extend(struct.pack("!B", h2|mask_bit))
        if self.mask is not None:
            nd.extend(self.mask)
        if self.data is not None:
            nd.extend(self.data)
        return nd

    @classmethod
    def client_ping(cls, data: bytes, mask: bytes = None) -> 'WsFrame':
        if mask is None:
            mask = bytes.fromhex('00 00 00 00')
        return WsFrame(opcode=WsFrame.PING, fin=True, mask=mask, data=data)

    @classmethod
    def client_close(cls, code: int, reason: str = None,
                     mask: bytes = None) -> 'WsFrame':
        data = bytearray(struct.pack("!H", code))
        if reason is not None:
            data.extend(reason.encode())
        if mask is None:
            mask = bytes.fromhex('00 00 00 00')
        return WsFrame(opcode=WsFrame.CLOSE, fin=True, mask=mask, data=data)


class WsFrameReader:

    def __init__(self, data: bytes):
        self.data = data

    def _read(self, n: int):
        if len(self.data) < n:
            raise EOFError(f'have {len(self.data)} bytes left, but {n} requested')
        elif n == 0:
            return b''
        chunk = self.data[:n]
        del self.data[:n]
        return chunk

    def next_frame(self):
        data = self._read(2)
        h1, h2 = struct.unpack("!BB", data)
        log.debug(f'parsed h1={h1} h2={h2} from {data}')
        fin = True if h1 & 0x80 else False
        opcode = h1 & 0xf
        has_mask = True if h2 & 0x80 else False
        mask = None
        dlen = h2 & 0x7f
        if dlen == 126:
            (dlen,) = struct.unpack("!H", self._read(2))
        elif dlen == 127:
            (dlen,) = struct.unpack("!Q", self._read(8))
        if has_mask:
            mask = self._read(4)
        return WsFrame(opcode=opcode, fin=fin, mask=mask, data=self._read(dlen))

    def eof(self):
        return len(self.data) == 0

    @classmethod
    def parse(cls, data: bytes):
        frames = []
        reader = WsFrameReader(data=data)
        while not reader.eof():
            frames.append(reader.next_frame())
        return frames



import inspect
import logging
import os
import subprocess
from shutil import copyfile
from typing import Dict, Any

from pyhttpd.certs import CertificateSpec
from pyhttpd.conf import HttpdConf
from pyhttpd.env import HttpdTestEnv, HttpdTestSetup

log = logging.getLogger(__name__)


class H2TestSetup(HttpdTestSetup):

    def __init__(self, env: 'HttpdTestEnv'):
        super().__init__(env=env)
        self.add_source_dir(os.path.dirname(inspect.getfile(H2TestSetup)))
        self.add_modules(["http2", "proxy_http2", "cgid", "autoindex", "ssl", "include"])

    def make(self):
        super().make()
        self._add_h2test()
        self._setup_data_1k_1m()

    def _add_h2test(self):
        local_dir = os.path.dirname(inspect.getfile(H2TestSetup))
        p = subprocess.run([self.env.apxs, '-c', 'mod_h2test.c'],
                           capture_output=True,
                           cwd=os.path.join(local_dir, 'mod_h2test'))
        rv = p.returncode
        if rv != 0:
            log.error(f"compiling md_h2test failed: {p.stderr}")
            raise Exception(f"compiling md_h2test failed: {p.stderr}")

        modules_conf = os.path.join(self.env.server_dir, 'conf/modules.conf')
        with open(modules_conf, 'a') as fd:
            # load our test module which is not installed
            fd.write(f"LoadModule h2test_module   \"{local_dir}/mod_h2test/.libs/mod_h2test.so\"\n")

    def _setup_data_1k_1m(self):
        s90 = "01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678\n"
        with open(os.path.join(self.env.gen_dir, "data-1k"), 'w') as f:
            for i in range(10):
                f.write(f"{i:09d}-{s90}")
        with open(os.path.join(self.env.gen_dir, "data-10k"), 'w') as f:
            for i in range(100):
                f.write(f"{i:09d}-{s90}")
        with open(os.path.join(self.env.gen_dir, "data-100k"), 'w') as f:
            for i in range(1000):
                f.write(f"{i:09d}-{s90}")
        with open(os.path.join(self.env.gen_dir, "data-1m"), 'w') as f:
            for i in range(10000):
                f.write(f"{i:09d}-{s90}")
        test1_docs = os.path.join(self.env.server_docs_dir, 'test1')
        self.env.mkpath(test1_docs)
        for fname in ["data-1k", "data-10k", "data-100k", "data-1m"]:
            src = os.path.join(self.env.gen_dir, fname)
            dest = os.path.join(test1_docs, fname)
            copyfile(src, dest)


class H2TestEnv(HttpdTestEnv):

    @classmethod
    @property
    def is_unsupported(cls):
        mpm_module = f"mpm_{os.environ['MPM']}" if 'MPM' in os.environ else 'mpm_event'
        return mpm_module == 'mpm_prefork'

    def __init__(self, pytestconfig=None):
        super().__init__(pytestconfig=pytestconfig)
        self.add_httpd_conf([
            "H2MinWorkers 1",
            "H2MaxWorkers 64",
            "Protocols h2 http/1.1 h2c",
        ])
        self.add_httpd_log_modules(["http2", "proxy_http2", "h2test", "proxy", "proxy_http"])
        self.add_cert_specs([
            CertificateSpec(domains=[
                f"push.{self._http_tld}",
                f"hints.{self._http_tld}",
                f"ssl.{self._http_tld}",
                f"pad0.{self._http_tld}",
                f"pad1.{self._http_tld}",
                f"pad2.{self._http_tld}",
                f"pad3.{self._http_tld}",
                f"pad8.{self._http_tld}",
            ]),
            CertificateSpec(domains=[f"noh2.{self.http_tld}"], key_type='rsa2048'),
        ])

    def setup_httpd(self, setup: HttpdTestSetup = None):
        super().setup_httpd(setup=H2TestSetup(env=self))


class H2Conf(HttpdConf):

    def __init__(self, env: HttpdTestEnv, extras: Dict[str, Any] = None):
        super().__init__(env=env, extras=HttpdConf.merge_extras(extras, {
            f"cgi.{env.http_tld}": [
                "SSLOptions +StdEnvVars",
                "AddHandler cgi-script .py",
                "<Location \"/h2test/echo\">",
                "    SetHandler h2test-echo",
                "</Location>",
                "<Location \"/h2test/delay\">",
                "    SetHandler h2test-delay",
                "</Location>",
                "<Location \"/h2test/error\">",
                "    SetHandler h2test-error",
                "</Location>",
            ]
        }))

    def start_vhost(self, domains, port=None, doc_root="htdocs", with_ssl=None,
                    ssl_module=None, with_certificates=None):
        super().start_vhost(domains=domains, port=port, doc_root=doc_root,
                            with_ssl=with_ssl, ssl_module=ssl_module,
                            with_certificates=with_certificates)
        if f"noh2.{self.env.http_tld}" in domains:
            protos = ["http/1.1"]
        elif port == self.env.https_port or with_ssl is True:
            protos = ["h2", "http/1.1"]
        else:
            protos = ["h2c", "http/1.1"]
        if f"test2.{self.env.http_tld}" in domains:
            protos = reversed(protos)
        self.add(f"Protocols {' '.join(protos)}")
        return self

    def add_vhost_noh2(self):
        domains = [f"noh2.{self.env.http_tld}", f"noh2-alias.{self.env.http_tld}"]
        self.start_vhost(domains=domains, port=self.env.https_port, doc_root="htdocs/noh2")
        self.add(["Protocols http/1.1", "SSLOptions +StdEnvVars"])
        self.end_vhost()
        self.start_vhost(domains=domains, port=self.env.http_port, doc_root="htdocs/noh2")
        self.add(["Protocols http/1.1", "SSLOptions +StdEnvVars"])
        self.end_vhost()
        return self

    def add_vhost_test1(self, proxy_self=False, h2proxy_self=False):
        return super().add_vhost_test1(proxy_self=proxy_self, h2proxy_self=h2proxy_self)

    def add_vhost_test2(self):
        return super().add_vhost_test2()



import inspect
import os
import re
import pytest

from .env import H2Conf, H2TestEnv


@pytest.mark.skipif(condition=H2TestEnv.is_unsupported, reason="mod_http2 not supported here")
class TestProxy:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        H2Conf(env).add_vhost_cgi(proxy_self=True).install()
        assert env.apache_restart() == 0

    def local_src(self, fname):
        return os.path.join(os.path.dirname(inspect.getfile(TestProxy)), fname)

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    def test_h2_500_01(self, env):
        url = env.mkurl("https", "cgi", "/proxy/hello.py")
        r = env.curl_get(url, 5)
        assert r.response["status"] == 200
        assert "HTTP/1.1" == r.response["json"]["protocol"]
        assert r.response["json"]["https"] == ""
        assert r.response["json"]["ssl_protocol"] == ""
        assert r.response["json"]["h2"] == ""
        assert r.response["json"]["h2push"] == ""

    # upload and GET again using curl, compare to original content
    def curl_upload_and_verify(self, env, fname, options=None):
        url = env.mkurl("https", "cgi", "/proxy/upload.py")
        fpath = os.path.join(env.gen_dir, fname)
        r = env.curl_upload(url, fpath, options=options)
        assert r.exit_code == 0
        assert 200 <= r.response["status"] < 300

        # why is the scheme wrong?
        r2 = env.curl_get(re.sub(r'http:', 'https:', r.response["header"]["location"]))
        assert r2.exit_code == 0
        assert r2.response["status"] == 200
        with open(self.local_src(fpath), mode='rb') as file:
            src = file.read()
        assert r2.response["body"] == src

    @pytest.mark.parametrize("name", [
        "data-1k", "data-10k", "data-100k", "data-1m",
    ])
    def test_h2_500_10(self, env, name, repeat):
        self.curl_upload_and_verify(env, name, ["--http2"])

    def test_h2_500_11(self, env):
        self.curl_upload_and_verify(env, "data-1k", [
            "--http1.1", "-H", "Content-Length:", "-H", "Transfer-Encoding: chunked"
        ])
        self.curl_upload_and_verify(env, "data-1k", ["--http2", "-H", "Content-Length:"])

    # POST some data using nghttp and see it echo'ed properly back
    def nghttp_post_and_verify(self, env, fname, options=None):
        url = env.mkurl("https", "cgi", "/proxy/echo.py")
        fpath = os.path.join(env.gen_dir, fname)
        r = env.nghttp().upload(url, fpath, options=options)
        assert r.exit_code == 0
        assert 200 <= r.response["status"] < 300
        with open(self.local_src(fpath), mode='rb') as file:
            src = file.read()
        if r.response["body"] != src:
            with open(os.path.join(env.gen_dir, "nghttp.out"), 'w') as fd:
                fd.write(r.outraw.decode())
                fd.write("\nstderr:\n")
                fd.write(r.stderr)
            assert r.response["body"] == src

    @pytest.mark.parametrize("name", [
        "data-1k", "data-10k", "data-100k", "data-1m",
    ])
    def test_h2_500_20(self, env, name, repeat):
        self.nghttp_post_and_verify(env, name, [])

    @pytest.mark.parametrize("name", [
        "data-1k", "data-10k", "data-100k", "data-1m",
    ])
    def test_h2_500_21(self, env, name, repeat):
        self.nghttp_post_and_verify(env, name, ["--no-content-length"])

    # upload and GET again using nghttp, compare to original content
    def nghttp_upload_and_verify(self, env, fname, options=None):
        url = env.mkurl("https", "cgi", "/proxy/upload.py")
        fpath = os.path.join(env.gen_dir, fname)

        r = env.nghttp().upload_file(url, fpath, options=options)
        assert r.exit_code == 0
        assert 200 <= r.response["status"] < 300
        assert r.response["header"]["location"]

        # why is the scheme wrong?
        r2 = env.nghttp().get(re.sub(r'http:', 'https:', r.response["header"]["location"]))
        assert r2.exit_code == 0
        assert r2.response["status"] == 200
        with open(self.local_src(fpath), mode='rb') as file:
            src = file.read()
        assert src == r2.response["body"]

    @pytest.mark.parametrize("name", [
        "data-1k", "data-10k", "data-100k", "data-1m",
    ])
    def test_h2_500_22(self, env, name):
        self.nghttp_upload_and_verify(env, name, [])

    @pytest.mark.parametrize("name", [
        "data-1k", "data-10k", "data-100k", "data-1m",
    ])
    def test_h2_500_23(self, env, name):
        self.nghttp_upload_and_verify(env, name, ["--no-content-length"])

    # upload using nghttp and check returned status
    def nghttp_upload_stat(self, env, fname, options=None):
        url = env.mkurl("https", "cgi", "/proxy/upload.py")
        fpath = os.path.join(env.gen_dir, fname)

        r = env.nghttp().upload_file(url, fpath, options=options)
        assert r.exit_code == 0
        assert 200 <= r.response["status"] < 300
        assert r.response["header"]["location"]

    def test_h2_500_24(self, env):
        for i in range(50):
            self.nghttp_upload_stat(env, "data-1k", ["--no-content-length"])

    # lets do some error tests
    def test_h2_500_30(self, env):
        url = env.mkurl("https", "cgi", "/proxy/h2test/error?status=500")
        r = env.curl_get(url)
        assert r.exit_code == 0, r
        assert r.response['status'] == 500
        url = env.mkurl("https", "cgi", "/proxy/h2test/error?error=timeout")
        r = env.curl_get(url)
        assert r.exit_code == 0, r
        assert r.response['status'] == 408

    # produce an error during response body
    def test_h2_500_31(self, env, repeat):
        url = env.mkurl("https", "cgi", "/proxy/h2test/error?body_error=timeout")
        r = env.curl_get(url)
        assert r.exit_code != 0, r
        #
        env.httpd_error_log.ignore_recent(
            lognos = [
                "AH01110"   # Network error reading response
            ]
        )

    # produce an error, fail to generate an error bucket
    def test_h2_500_32(self, env, repeat):
        url = env.mkurl("https", "cgi", "/proxy/h2test/error?body_error=timeout&error_bucket=0")
        r = env.curl_get(url)
        assert r.exit_code != 0, r
        #
        env.httpd_error_log.ignore_recent(
            lognos = [
                "AH01110"   # Network error reading response
            ]
        )



import logging
import os

import pytest
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '../..'))

from .env import H2TestEnv


def pytest_report_header(config, startdir):
    env = H2TestEnv()
    return f"mod_h2 [apache: {env.get_httpd_version()}, mpm: {env.mpm_module}, {env.prefix}]"


@pytest.fixture(scope="package")
def env(pytestconfig) -> H2TestEnv:
    level = logging.INFO
    console = logging.StreamHandler()
    console.setLevel(level)
    console.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    logging.getLogger('').addHandler(console)
    logging.getLogger('').setLevel(level=level)
    env = H2TestEnv(pytestconfig=pytestconfig)
    env.setup_httpd()
    env.apache_access_log_clear()
    env.httpd_error_log.clear_log()
    return env


@pytest.fixture(autouse=True, scope="package")
def _h2_package_scope(env):
    env.httpd_error_log.add_ignored_lognos([
        'AH10400',  # warning that 'enablereuse' has not effect in certain configs
        'AH00045',  # child did not exit in time, SIGTERM was sent
    ])
    yield
    assert env.apache_stop() == 0



import os
import re
import socket
from threading import Thread

import pytest

from pyhttpd.conf import HttpdConf
from pyhttpd.result import ExecResult


class UDSFaker:

    def __init__(self, path):
        self._uds_path = path
        self._done = False

    def start(self):
        def process(self):
            self._socket.listen(1)
            self._process()

        try:
            os.unlink(self._uds_path)
        except OSError:
            if os.path.exists(self._uds_path):
                raise
        self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._socket.bind(self._uds_path)
        self._thread = Thread(target=process, daemon=True, args=[self])
        self._thread.start()

    def stop(self):
        self._done = True
        self._socket.close()

    def _process(self):
        while self._done is False:
            try:
                c, client_address = self._socket.accept()
                try:
                    data = c.recv(16)
                    c.sendall("""HTTP/1.1 200 Ok
Server: UdsFaker
Content-Type: application/json
Content-Length: 19

{ "host": "faked" }""".encode())
                finally:
                    c.close()

            except ConnectionAbortedError:
                self._done = True


class TestProxyUds:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        # setup 3 vhosts on https: for reverse, forward and
        # mixed proxying to a unix: domain socket
        # We setup a UDSFaker running that returns a fixed response
        UDS_PATH = f"{env.gen_dir}/proxy_02.sock"
        TestProxyUds.UDS_PATH = UDS_PATH
        faker = UDSFaker(path=UDS_PATH)
        faker.start()

        conf = HttpdConf(env)
        conf.add("ProxyPreserveHost on")
        conf.start_vhost(domains=[env.d_reverse], port=env.https_port)
        conf.add([
            f"ProxyPass / unix:{UDS_PATH}|http://127.0.0.1:{env.http_port}/"
        ])
        conf.end_vhost()

        conf.start_vhost(domains=[env.d_forward], port=env.https_port)
        conf.add([
            "ProxyRequests on"
        ])
        conf.end_vhost()

        conf.start_vhost(domains=[env.d_mixed], port=env.https_port)
        conf.add([
            f"ProxyPass / unix:{UDS_PATH}|http://127.0.0.1:{env.http_port}/",
            "ProxyRequests on"
        ])
        conf.end_vhost()
        conf.install()
        assert env.apache_restart() == 0
        yield
        faker.stop()

    @pytest.mark.parametrize(["via", "seen"], [
        ["reverse", "faked"],
        ["mixed", "faked"],
    ])
    def test_proxy_02_001(self, env, via, seen):
        # make requests to a reverse proxy https: vhost to the http: vhost
        # check that we see the document we expect there (host matching worked)
        r = env.curl_get(f"https://{via}.{env.http_tld}:{env.https_port}/alive.json", 5)
        assert r.response["status"] == 200
        assert r.json['host'] == seen

    @pytest.mark.parametrize(["via", "seen"], [
        ["forward", "generic"],
        ["mixed", "faked"],
    ])
    def test_proxy_02_002(self, env, via, seen):
        # make requests to a forward proxy https: vhost to the http: vhost
        # check that we see the document we expect there (host matching worked)
        # we need to explicitly provide a Host: header since mod_proxy cannot
        # resolve the name via DNS.
        if not env.curl_is_at_least('8.0.0'):
            pytest.skip(f'need at least curl v8.0.0 for this')
        domain = f"{via}.{env.http_tld}"
        r = env.curl_get(f"http://127.0.0.1:{env.http_port}/alive.json", 5, options=[
            '-H', f"Host: {domain}",
            '--proxy', f"https://{domain}:{env.https_port}/",
            '--resolve', f"{domain}:{env.https_port}:127.0.0.1",
            '--proxy-cacert', f"{env.get_ca_pem_file(domain)}",

        ])
        assert r.exit_code == 0, f"{r.stdout}{r.stderr}"
        assert r.response["status"] == 200
        assert r.json['host'] == seen

    @pytest.mark.parametrize(["via", "exp_status"], [
        ["reverse", 400],
        ["forward", 500],
        ["mixed", 500],
    ])
    def test_proxy_02_003(self, env, via, exp_status):
        # make requests to a forward proxy https: vhost and GET
        # a URL which carries the unix: domain socket.
        # This needs to fail.
        domain = f"{via}.{env.http_tld}"
        r = env.run(args=[
            'openssl', 's_client', '-connect', f"127.0.0.1:{env.https_port}",
            '-servername', domain,
            '-crlf', '-ign_eof',
            '-CAfile', env.get_ca_pem_file(domain)
        ], intext=f"""GET unix:{TestProxyUds.UDS_PATH}|http://127.0.0.1:{env.http_port}/alive.json HTTP/1.1
Host: {domain}

""")
        assert r.exit_code == 0, f"{r.stdout}{r.stderr}"
        lines = r.stdout.split('\n')
        rlines = None
        for idx, l in enumerate(lines):
            if l.startswith('HTTP/'):
                rlines = lines[idx:]
        assert rlines, f"No response found in: {r.stdout}"
        r2 = self.parse_response(rlines)
        assert r2.response
        assert r2.response['status'] == exp_status
        #
        env.httpd_error_log.ignore_recent(
            lognos = [
                "AH01144"   #  No protocol handler was valid for the URL
            ]
        )

    def parse_response(self, lines) -> ExecResult:
        exp_body = False
        exp_stat = True
        r = ExecResult(args=[], exit_code=0, stdout=b'', stderr=b'')
        header = {}
        body = []
        for line in lines:
            if exp_stat:
                m = re.match(r'^(\S+) (\d+) (.*)$', line)
                assert m, f"first line no HTTP status line: {line}"
                r.add_response({
                    "protocol": m.group(1),
                    "status": int(m.group(2)),
                    "description": m.group(3),
                    "body": r.outraw
                })
                header = {}
                exp_stat = False
                exp_body = False
            elif re.match(r'^\r?$', line):
                exp_body = True
            elif exp_body:
                body.append(line)
            else:
                m = re.match(r'^([^:]+):\s*(.*)$', line)
                assert m, f"not a header line: {line}"
                header[m.group(1).lower()] = m.group(2)
        if r.response:
            r.response["header"] = header
            r.response["body"] = body
        return r



#!/usr/bin/env python3
import os, sys
from requestparser import get_request_params


forms, files = get_request_params()

status = '200 Ok'

try:
    ename = forms['name']

    # Test if the file was uploaded
    if ename is not None:
        val = os.environ[ename] if ename in os.environ else ""
        print("Status: 200")
        print("""\
Content-Type: text/plain\n""")
        print(f"{ename}={val}")

    else:
        print("Status: 400 Parameter Missing")
        print("""\
Content-Type: text/html\n
    <html><body>
    <p>No name was specified: name</p>
    </body></html>""")

except KeyError:
    print("Status: 200 Ok")
    print("""\
Content-Type: text/html\n
    <html><body>
    Echo <form method="POST" enctype="application/x-www-form-urlencoded">
    <input type="text" name="name">
    <button type="submit">submit</button></form>
    </body></html>""")
    pass






import pytest

from .env import H2Conf, H2TestEnv


@pytest.mark.skipif(condition=H2TestEnv.is_unsupported, reason="mod_http2 not supported here")
@pytest.mark.skipif(not H2TestEnv().h2load_is_at_least('1.41.0'),
                    reason="h2load misses --connect-to option")
class TestLoadGet:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        H2Conf(env).add_vhost_cgi().add_vhost_test1().install()
        assert env.apache_restart() == 0

    def check_h2load_ok(self, env, r, n):
        assert 0 == r.exit_code
        r = env.h2load_status(r)
        assert n == r.results["h2load"]["requests"]["total"], f'{r.results}'
        assert n == r.results["h2load"]["requests"]["started"], f'{r.results}'
        assert n == r.results["h2load"]["requests"]["done"], f'{r.results}'
        assert n == r.results["h2load"]["requests"]["succeeded"], f'{r.results}'
        assert n == r.results["h2load"]["status"]["2xx"], f'{r.results}'
        assert 0 == r.results["h2load"]["status"]["3xx"], f'{r.results}'
        assert 0 == r.results["h2load"]["status"]["4xx"], f'{r.results}'
        assert 0 == r.results["h2load"]["status"]["5xx"], f'{r.results}'
    
    # test load on cgi script, single connection, different sizes
    @pytest.mark.parametrize("start", [
        1000, 80000
    ])
    def test_h2_700_10(self, env, start):
        assert env.is_live()
        text = "X"
        chunk = 32
        for n in range(0, 5):
            args = [env.h2load, "-n", "%d" % chunk, "-c", "1", "-m", "10",
                    f"--connect-to=localhost:{env.https_port}",
                    f"--base-uri={env.mkurl('https', 'cgi', '/')}",
            ]
            for i in range(0, chunk):
                args.append(env.mkurl("https", "cgi", ("/mnot164.py?count=%d&text=%s" % (start+(n*chunk)+i, text))))
            r = env.run(args)
            self.check_h2load_ok(env, r, chunk)

    # test load on cgi script, single connection
    @pytest.mark.parametrize("conns", [
        1, 2, 16
    ])
    def test_h2_700_11(self, env, conns):
        assert env.is_live()
        text = "X"
        start = 1200
        chunk = 64
        for n in range(0, 5):
            args = [env.h2load, "-n", "%d" % chunk, "-c", "%d" % conns, "-m", "10",
                    f"--connect-to=localhost:{env.https_port}",
                    f"--base-uri={env.mkurl('https', 'cgi', '/')}",
            ]
            for i in range(0, chunk):
                args.append(env.mkurl("https", "cgi", ("/mnot164.py?count=%d&text=%s" % (start+(n*chunk)+i, text))))
            r = env.run(args)
            self.check_h2load_ok(env, r, chunk)

    # test window sizes, connection and stream
    @pytest.mark.parametrize("connbits,streambits", [
        [10, 16],  # 1k connection window, 64k stream windows
        [10, 30],  # 1k connection window, huge stream windows
        [30, 8],  # huge conn window, 256 bytes stream windows
    ])
    @pytest.mark.skip('awaiting mpm_event improvements')
    def test_h2_700_20(self, env, connbits, streambits):
        if not env.httpd_is_at_least("2.5.0"):
            pytest.skip(f'need at least httpd 2.5.0 for this')
        conf = H2Conf(env, extras={
            'base': [
                'StartServers 1',
            ]
        })
        conf.add_vhost_cgi().add_vhost_test1().install()
        assert env.apache_restart() == 0
        assert env.is_live()
        n = 2000
        conns = 50
        parallel = 10
        args = [
            env.h2load,
            '-n', f'{n}', '-t', '1',
            '-c', f'{conns}', '-m', f'{parallel}',
            '-W', f'{connbits}',  # connection window bits
            '-w', f'{streambits}',  # stream window bits
            f'--connect-to=localhost:{env.https_port}',
            f'--base-uri={env.mkurl("https", "test1", "/")}',
            "/data-100k"
        ]
        r = env.run(args)
        self.check_h2load_ok(env, r, n)


import inspect
import logging
import os
import subprocess
from typing import Dict, Any

from pyhttpd.certs import CertificateSpec
from pyhttpd.conf import HttpdConf
from pyhttpd.env import HttpdTestEnv, HttpdTestSetup

log = logging.getLogger(__name__)


class ProxyTestSetup(HttpdTestSetup):

    def __init__(self, env: 'HttpdTestEnv'):
        super().__init__(env=env)
        self.add_source_dir(os.path.dirname(inspect.getfile(ProxyTestSetup)))
        self.add_modules(["proxy", "proxy_http", "proxy_balancer", "lbmethod_byrequests"])


class ProxyTestEnv(HttpdTestEnv):

    def __init__(self, pytestconfig=None):
        super().__init__(pytestconfig=pytestconfig)
        self.add_httpd_conf([
                         ])
        self._d_reverse = f"reverse.{self.http_tld}"
        self._d_forward = f"forward.{self.http_tld}"
        self._d_mixed = f"mixed.{self.http_tld}"

        self.add_httpd_log_modules(["proxy", "proxy_http", "proxy_balancer", "lbmethod_byrequests", "ssl"])
        self.add_cert_specs([
            CertificateSpec(domains=[
                self._d_forward, self._d_reverse, self._d_mixed
            ]),
            CertificateSpec(domains=[f"noh2.{self.http_tld}"], key_type='rsa2048'),
        ])

    def setup_httpd(self, setup: HttpdTestSetup = None):
        super().setup_httpd(setup=ProxyTestSetup(env=self))

    @property
    def d_forward(self):
        return self._d_forward

    @property
    def d_reverse(self):
        return self._d_reverse

    @property
    def d_mixed(self):
        return self._d_mixed



#!/usr/bin/env python3
import argparse
import asyncio
import logging
import os
import sys
import time

import websockets.server as ws_server
from websockets.exceptions import ConnectionClosedError

log = logging.getLogger(__name__)

logging.basicConfig(
    format="[%(asctime)s] %(message)s",
    level=logging.DEBUG,
)


async def echo(websocket):
    try:
        async for message in websocket:
            try:
                log.info(f'got request {message}')
            except Exception as e:
                log.error(f'error {e} getting path from {message}')
            await websocket.send(message)
    except ConnectionClosedError:
        pass


async def on_async_conn(conn):
    rpath = str(conn.path)
    pcomps = rpath[1:].split('/')
    if len(pcomps) == 0:
        pcomps = ['echo']  # default handler
    log.info(f'connection for {pcomps}')
    if pcomps[0] == 'echo':
        log.info(f'/echo endpoint')
        for message in await conn.recv():
            await conn.send(message)
    elif pcomps[0] == 'text':
        await conn.send('hello!')
    elif pcomps[0] == 'file':
        if len(pcomps) < 2:
            conn.close(code=4999, reason='unknown file')
            return
        fpath = os.path.join('../', pcomps[1])
        if not os.path.exists(fpath):
            conn.close(code=4999, reason='file not found')
            return
        bufsize = 0
        if len(pcomps) > 2:
            bufsize = int(pcomps[2])
        if bufsize <= 0:
            bufsize = 16*1024
        delay_ms = 0
        if len(pcomps) > 3:
            delay_ms = int(pcomps[3])
        n = 1
        if len(pcomps) > 4:
            n = int(pcomps[4])
        for _ in range(n):
            with open(fpath, 'r+b') as fd:
                while True:
                    buf = fd.read(bufsize)
                    if buf is None or len(buf) == 0:
                        break
                    await conn.send(buf)
                    if delay_ms > 0:
                        time.sleep(delay_ms/1000)
    else:
        log.info(f'unknown endpoint: {rpath}')
        await conn.close(code=4999, reason='path unknown')
    await conn.close(code=1000, reason='')


async def run_server(port):
    log.info(f'starting server on port {port}')
    async with ws_server.serve(ws_handler=on_async_conn,
                               host="localhost", port=port):
        await asyncio.Future()


async def main():
    parser = argparse.ArgumentParser(prog='scorecard',
                                     description="Run a websocket echo server.")
    parser.add_argument("--port", type=int,
                        default=0, help="port to listen on")
    args = parser.parse_args()

    if args.port == 0:
        sys.stderr.write('need --port\n')
        sys.exit(1)

    logging.basicConfig(
        format="%(asctime)s %(message)s",
        level=logging.DEBUG,
    )
    await run_server(args.port)


if __name__ == "__main__":
    asyncio.run(main())



import os
import pytest

from .env import H1Conf


# The trailer tests depend on "nghttp" as no other client seems to be able to send those
# rare things.
class TestTrailers:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        H1Conf(env).add_vhost_cgi(proxy_self=True).install()
        assert env.apache_restart() == 0

    # check that we get a trailer out when telling the handler to add one
    def test_h1_005_01(self, env):
        if not env.httpd_is_at_least("2.5.0"):
            pytest.skip(f'need at least httpd 2.5.0 for this')
        url = env.mkurl("https", "cgi", "/h1test/echo")
        host = f"cgi.{env.http_tld}"
        fpath = os.path.join(env.gen_dir, "data-1k")
        r = env.curl_upload(url, fpath, options=["--header", "Add-Trailer: 005_01"])
        assert r.exit_code == 0, f"{r}"
        assert 200 <= r.response["status"] < 300
        assert r.response["trailer"], f"no trailers received: {r}"
        assert "h1test-add-trailer" in r.response["trailer"]
        assert r.response["trailer"]["h1test-add-trailer"] == "005_01"

    # check that we get out trailers through the proxy
    def test_h1_005_02(self, env):
        if not env.httpd_is_at_least("2.5.0"):
            pytest.skip(f'need at least httpd 2.5.0 for this')
        url = env.mkurl("https", "cgi", "/proxy/h1test/echo")
        host = f"cgi.{env.http_tld}"
        fpath = os.path.join(env.gen_dir, "data-1k")
        r = env.curl_upload(url, fpath, options=["--header", "Add-Trailer: 005_01"])
        assert r.exit_code == 0, f"{r}"
        assert 200 <= r.response["status"] < 300
        assert r.response["trailer"], f"no trailers received: {r}"
        assert "h1test-add-trailer" in r.response["trailer"]
        assert r.response["trailer"]["h1test-add-trailer"] == "005_01"



import re
import socket
from typing import List, Optional

import pytest

from .env import H1Conf

class TestRequestUnsafe:

    @pytest.fixture(autouse=True, scope='class')
    def _class_scope(self, env):
        conf = H1Conf(env)
        conf.add([
            "HttpProtocolOptions Unsafe",
        ])
        conf.install()
        assert env.apache_restart() == 0

    # unsafe tests from t/apache/http_strict.t
    # possible expected results:
    #   0:       any HTTP error
    #   1:       any HTTP success
    #   200-500: specific HTTP status code
    #   None:   HTTPD should drop connection without error message
    @pytest.mark.parametrize(["intext", "status", "lognos"], [
        ["GET / HTTP/1.0\r\n\r\n", 1, None],
        ["GET / HTTP/1.0\n\n", 1, None],
        ["get / HTTP/1.0\r\n\r\n", 501, ["AH00135"]],
        ["G ET / HTTP/1.0\r\n\r\n", 400, None],
        ["G\0ET / HTTP/1.0\r\n\r\n", 400, None],
        ["G/T / HTTP/1.0\r\n\r\n", 501, ["AH00135"]],
        ["GET /\0 HTTP/1.0\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\0\r\n\r\n", 400, None],
        ["GET\f/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET\r/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET\t/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET / HTT/1.0\r\n\r\n", 0, None],
        ["GET / HTTP/1.0\r\nHost: localhost\r\n\r\n", 1, None],
        ["GET / HTTP/2.0\r\nHost: localhost\r\n\r\n", 1, None],
        ["GET / HTTP/1.2\r\nHost: localhost\r\n\r\n", 1, None],
        ["GET / HTTP/1.11\r\nHost: localhost\r\n\r\n", 400, None],
        ["GET / HTTP/10.0\r\nHost: localhost\r\n\r\n", 400, None],
        ["GET / HTTP/1.0  \r\nHost: localhost\r\n\r\n", 200, None],
        ["GET / HTTP/1.0 x\r\nHost: localhost\r\n\r\n", 400, None],
        ["GET / HTTP/\r\nHost: localhost\r\n\r\n", 0, None],
        ["GET / HTTP/0.9\r\n\r\n", 0, None],
        ["GET / HTTP/0.8\r\n\r\n", 0, None],
        ["GET /\x01 HTTP/1.0\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nFoo: bar\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nFoo:bar\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nFoo: b\0ar\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nFoo: b\x01ar\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nFoo\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nFoo bar\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\n: bar\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nX: bar\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nFoo bar:bash\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nFoo :bar\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\n Foo:bar\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nF\x01o: bar\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nF\ro: bar\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nF\to: bar\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nFo: b\tar\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nFo: bar\r\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\r", None, None],
        ["GET /\r\n", 0, None],
        ["GET /#frag HTTP/1.0\r\n", 400, None],
        ["GET / HTTP/1.0\r\nHost: localhost\r\nHost: localhost\r\n\r\n", 200, None],
        ["GET http://017700000001/ HTTP/1.0\r\n\r\n", 200, None],
        ["GET http://0x7f.1/ HTTP/1.0\r\n\r\n", 200, None],
        ["GET http://127.0.0.1/ HTTP/1.0\r\n\r\n", 200, None],
        ["GET http://127.01.0.1/ HTTP/1.0\r\n\r\n", 200, None],
        ["GET http://%3127.0.0.1/ HTTP/1.0\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nHost: localhost:80\r\nHost: localhost:80\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nHost: localhost:80 x\r\n\r", 400, None],
        ["GET http://localhost:80/ HTTP/1.0\r\n\r\n", 200, None],
        ["GET http://localhost:80x/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET http://localhost:80:80/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET http://localhost::80/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET http://foo@localhost:80/ HTTP/1.0\r\n\r\n", 200, None],
        ["GET http://[::1]/ HTTP/1.0\r\n\r\n", 1, None],
        ["GET http://[::1:2]/ HTTP/1.0\r\n\r\n", 1, None],
        ["GET http://[4712::abcd]/ HTTP/1.0\r\n\r\n", 1, None],
        ["GET http://[4712::abcd:1]/ HTTP/1.0\r\n\r\n", 1, None],
        ["GET http://[4712::abcd::]/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET http://[4712:abcd::,]/ HTTP/1.0\r\n\r\n", 1, None],
        ["GET http://[4712::abcd]:8000/ HTTP/1.0\r\n\r\n", 1, None],
        ["GET http://4713::abcd:8001/ HTTP/1.0\r\n\r\n", 400, None],
        ["GET / HTTP/1.0\r\nHost: [::1]\r\n\r\n", 1, None],
        ["GET / HTTP/1.0\r\nHost: [::1:2]\r\n\r\n", 1, None],
        ["GET / HTTP/1.0\r\nHost: [4711::abcd]\r\n\r\n", 1, None],
        ["GET / HTTP/1.0\r\nHost: [4711::abcd:1]\r\n\r\n", 1, None],
        ["GET / HTTP/1.0\r\nHost: [4711:abcd::]\r\n\r\n", 1, None],
        ["GET / HTTP/1.0\r\nHost: [4711::abcd]:8000\r\n\r\n", 1, None],
        ["GET / HTTP/1.0\r\nHost: 4714::abcd:8001\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nHost: abc\xa0\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nHost: abc\\foo\r\n\r\n", 400, None],
        ["GET http://foo/ HTTP/1.0\r\nHost: bar\r\n\r\n", 200, None],
        ["GET http://foo:81/ HTTP/1.0\r\nHost: bar\r\n\r\n", 200, None],
        ["GET http://[::1]:81/ HTTP/1.0\r\nHost: bar\r\n\r\n", 200, None],
        ["GET http://10.0.0.1:81/ HTTP/1.0\r\nHost: bar\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nHost: foo-bar.example.com\r\n\r\n", 200, None],
        ["GET / HTTP/1.0\r\nHost: foo_bar.example.com\r\n\r\n", 200, None],
        ["GET http://foo_bar/ HTTP/1.0\r\n\r\n", 200, None],
    ])
    def test_h1_006_01(self, env, intext, status: Optional[int], lognos: Optional[List[str]]):
        with socket.create_connection(('localhost', int(env.http_port))) as sock:
            # on some OS, the server does not see our connection until there is
            # something incoming
            sock.sendall(intext.encode())
            sock.shutdown(socket.SHUT_WR)
            buff = sock.recv(1024)
            msg = buff.decode()
            if status is None:
                assert len(msg) == 0, f"unexpected answer: {msg}"
            else:
                assert len(msg) > 0, "no answer from server"
                rlines = msg.splitlines()
                response = rlines[0]
                m = re.match(r'^HTTP/1.1 (\d+)\s+(\S+)', response)
                assert m or status == 0, f"unrecognized response: {rlines}"
                if status == 1:
                    assert int(m.group(1)) >= 200
                elif status == 0:
                    # headerless 0.9 response, yuk
                    assert len(rlines) >= 1, f"{rlines}"
                elif status > 0:
                    assert int(m.group(1)) == status, f"{rlines}"
                else:
                    assert int(m.group(1)) >= 400, f"{rlines}"
                #
                if lognos is not None:
                    env.httpd_error_log.ignore_recent(lognos = lognos)
